from __future__ import print_function
import argparse
import numpy as np

import torch
import torch.nn.functional as F
from torch.autograd import Variable
from torch import nn, optim
from torch.nn import functional as F
from torchvision import datasets, transforms
from torchvision.utils import save_image
from load_data import load_dataset
from numpy import loadtxt, savetxt


def sample_gumbel(shape, eps=1e-20):
    """Sample from Gumbel(0, 1)"""
    U = torch.rand(shape).cuda()
    return -Variable(torch.log(-torch.log(U + eps) + eps))

def gumbel_softmax_sample(logits, temperature):
    """ Draw a sample from the Gumbel-Softmax distribution"""
    y = logits + sample_gumbel(logits.size())
    
    return F.softmax(y / temperature, dim=-1)

def gumbel_softmax(logits, temperature, latent_dim, categorical_dim):
    """
    ST-gumple-softmax
    input: [*, n_class]
    return: flatten --> [*, n_class] an one-hot vector
    """
    
    y = gumbel_softmax_sample(logits, temperature)
    y_hard = y
    #print("y1 shape", y.shape)
    
    shape = y.size()
    _, ind = y.max(dim=-1)
    y_hard = torch.zeros_like(y).view(-1, shape[-1])
    y_hard.scatter_(1, ind.view(-1, 1), 1)
    y_hard = y_hard.view(*shape)
    y_hard = (y_hard - y).detach() + y    #the gradients are taken wrt to y
 #   print('yhard: {:11.6f}'.format(y_hard))
    y_hard = y_hard.view(-1,latent_dim*categorical_dim)
    
    return y_hard



class VAE_gumbel(nn.Module):

    def __init__(self,temp, latent_dim, categorical_dim):
        super().__init__()

        self.z_size = categorical_dim*latent_dim
        self.input_length = 80
        self.padding = 1
        self.categorical_dim = categorical_dim
        self.latent_dim = latent_dim
        
        self.conv1 = nn.Conv1d(in_channels=4, out_channels=12, kernel_size=3, padding=self.padding )
        self.pool = nn.MaxPool1d(kernel_size=2, stride=2, padding=1,  return_indices=True)
        self.pool2 = nn.MaxPool1d(kernel_size=2, stride=2, padding=0,  return_indices=True)
        self.conv2 = nn.Conv1d(in_channels=12, out_channels=8, kernel_size=3, padding=self.padding )
        
        #input length after one convolution and pooling layer
        self.input_length = ((self.input_length - 3 + 2*self.padding) // 1 + 1 - 2) // 2 + 1
        #input length after the second convolution and pooling layer
        self.input_length = ((self.input_length - 3 + 2*self.padding) // 1 + 1 - 2) // 2 + 1       
        self.dense_input = self.input_length*8
        
        
        self.fc1 = nn.Linear(self.dense_input, self.z_size)
        
        self.conv1_D = nn.ConvTranspose1d(in_channels=12, out_channels=4, kernel_size=3, padding=1)
      #  self.pool = nn.MaxPool1d(kernel_size=2, stride=2)
        self.unpool_D = nn.MaxUnpool1d(kernel_size=2, stride=2, padding=1)
        self.unpool2_D = nn.MaxUnpool1d(kernel_size=2, stride=2, padding=0)
        self.conv2_D = nn.ConvTranspose1d(in_channels=8, out_channels=12, kernel_size=3, padding=1)
        self.fc1_D = nn.Linear(self.z_size, self.dense_input)
        self.unflatten = nn.Unflatten(1, torch.Size([8, 30]))
        self.m = nn.Softmax(dim=1)
        
    def encode(self, x):

        x = F.LeakyReLU(self.conv1(x))
        size1 = x.size()
      #  print(x.shape) # shape [100, 16, 116]
        x, indices1 = self.pool2(x)    
      #  print(x.shape)  #shape [100, 16, 58]
        x = F.LeakyReLU(self.conv2(x))
      #  print(x.shape) # shape [100, 8, 56]
        size2 = x.size()
        x, indices2 = self.pool2(x)
        
      #  print(x.shape) #shape [100, 8, 28]
        x = torch.flatten(x, 1) 
      #  print(x.shape) #[100, 224]
        x = self.fc1(x)
        return x, indices1, indices2, size1, size2

    def decode(self, z, indices1, indices2):

        x = self.fc1_D(z)
       # print(x.shape) #shape [100, 224]
        x = self.unflatten(x)
       # print(x.shape)  # shape [100, 8, 28]
        x = self.unpool2_D(x, indices2)
       # print(x.shape) # shape [100, 25, 56] 
        x = F.relu(self.conv2_D(x))
       # print(x.shape)  # shape [100, 32, 58] 
        x = self.unpool2_D(x, indices1)
       # print(x.shape)  # shape [100, 32, 116] 
    
        
        x = F.relu(self.conv1_D(x))
    #    print(x.shape)
        x = self.m(x)
    #    print(x.shape)
    #    x = self.dropout(x)
        return x
   
    
    def forward(self, x, y_values=None, temp=0.7):
      
        q, indices1, indices2 = self.encode(x)#.view(-1, 480))
        q_y = q.view(q.size(0), self.latent_dim, self.categorical_dim)
        z = gumbel_softmax(q_y,temp, self.latent_dim, self.categorical_dim)
        return self.decode(z, indices1, indices2), F.softmax(q, dim=2)

class VAE_gumbel_hp_tuning(nn.Module):

    def __init__(self,temp_rate, latent_dim, categorical_dim, no_filters_conv1, no_filters_conv2, filter_size, filter2_size, stride,
                 pool_size, pool_stride, length, act, dropout):
        super().__init__()

        self.fc2_layer = 2 ############ 2:VANILLA VAE 1: GS VAE with 2 fc layer  0: GS VAE with 1 fc layer 
        self.num_conv_layers = 1#num_conv
        
        stride_conv_2 = (1,2)
        
        self.temp = 1.0
        self.temp_min = 1.0
        self.temp_rate = temp_rate
        self.hidden = 128
                
        self.height = 4
        self.length = length
        
        self.categorical_dim = categorical_dim
        self.latent_dim = latent_dim
        self.z_size = self.categorical_dim * self.latent_dim
        self.padding = (0,1)
        self.act = act
        
        
        self.conv1 = nn.Conv2d(in_channels=1, out_channels=no_filters_conv1, kernel_size=filter_size, padding=self.padding, stride=stride)
        self.pool1 = nn.MaxPool2d(kernel_size=pool_size, stride=pool_stride, return_indices=True)
        self.pool2 = nn.MaxPool2d(kernel_size=pool_size, stride=pool_stride, return_indices=True)
        self.conv2 = nn.Conv2d(in_channels=no_filters_conv1, out_channels=no_filters_conv2, kernel_size=filter2_size, padding=self.padding, stride=stride_conv_2)
        
        #input length after one convolution and pooling layer
        self.input_length = ((self.length - filter_size[1] + 2*self.padding[1]) // stride[1] + 1 - pool_size[1]) // pool_stride[1] + 1
        self.input_height = ((self.height - filter_size[0] + 2*self.padding[0]) // stride[0] + 1 - pool_size[0]) // pool_stride[0] + 1
        #input length after the second convolution and pooling layer
        if self.num_conv_layers == 2:
            self.input_length = (self.input_length - filter2_size[1] + 2*self.padding[1]) // stride_conv_2[1] + 1 #- pool_size[1]) // pool_stride[1] + 1   
            self.input_height = (self.input_height - filter2_size[0] + 2*self.padding[0]) // stride_conv_2[0] + 1 #- pool_size[0]) // pool_stride[0] + 1
        
        self.dense_input = self.input_length*no_filters_conv2*self.input_height
        

        if self.fc2_layer ==1:
            self.fc1 = nn.Linear(self.dense_input, int(self.dense_input/2)) 
            self.fc2 = nn.Linear(int(self.dense_input/2), self.z_size)
        elif self.fc2_layer==0:
            self.fc1 = nn.Linear(self.dense_input, self.z_size)
            
        elif self.fc2_layer==2:
            self.fc1 = nn.Linear(self.dense_input, self.hidden)
            self.encodinglayer2_mean = nn.Sequential(nn.Linear(self.hidden, self.z_size))
            self.encodinglayer2_logvar = nn.Sequential(nn.Linear(self.hidden, self.z_size))
            self.decodinglayer = nn.Sequential(
            nn.Linear(self.z_size, self.hidden),
            nn.ReLU(),
            )
            self.fc1_D = nn.Linear(self.hidden, self.dense_input) 
            
            
        self.conv1_D = nn.ConvTranspose2d(in_channels=no_filters_conv1, out_channels=1, kernel_size=filter_size, padding=self.padding, stride=stride)
        self.unpool_D = nn.MaxUnpool2d(kernel_size=pool_size, stride=pool_stride)
        self.unpool2_D = nn.MaxUnpool2d(kernel_size=pool_size, stride=pool_stride)
        if self.num_conv_layers == 2:
            self.conv2_D = nn.ConvTranspose2d(in_channels=no_filters_conv2, out_channels=no_filters_conv1, kernel_size=filter2_size, padding=self.padding, stride=stride_conv_2)
         
        self.unflatten = nn.Unflatten(1, torch.Size([no_filters_conv2, self.input_height,
                                                     int(self.dense_input/(no_filters_conv2*self.input_height))]))

     
        if self.act == 'softmax':
            self.m = nn.Softmax(dim=1)       
        elif self.act == 'sigmoid':
            self.m = nn.Sigmoid()
        
        if self.fc2_layer ==1:
            self.fc1_D = nn.Linear(int(self.dense_input/2), self.dense_input) 
            self.fc2_D = nn.Linear(self.z_size, int(self.dense_input/2))    
        elif self.fc2_layer ==0:
            self.fc1_D = nn.Linear(self.z_size, self.dense_input) 
            
        #self.bn1 = nn.BatchNorm2d(no_filters_conv1, eps=1e-05, momentum=0.1, affine=True)
     #   self.bn2 = nn.BatchNorm2d(no_filters_conv2, eps=1e-05, momentum=0.1, affine=True),
        self.md = nn.Dropout(p=dropout)
        
    def encode(self, x):
        indices2 = 0
        indices1 = 0
        size1 = 0
        size2 = 0        
        indices2 = 0
        size_conv1=0 
        size_conv2=0
        
        x = torch.unsqueeze(x, 1)
      #  print(x.shape)             #[256, 1, 4, 80]
        size_conv1 = x.shape
        x = self.conv1(x)
      #  print(x.shape) 
        x = F.relu(x)#self.bn1(x))
        size1 = x.size()
        x = self.md(x)                                                ############DROPOUT
      #  print(x.shape)          #[256, 8, 4, 38]
        x, indices1 = self.pool1(x)  
      #  print(x.shape) # [256, 8, 4, 19]
        if self.num_conv_layers == 2:
            size_conv2 = x.shape
            x = F.relu(self.conv2(x))    #[256, 8, 1, 9]
            x = self.md(x)                                                ############DROPOUT
      #      print(x.shape)
        size2 = x.size()  
        if self.num_conv_layers == 2:
            x=x
            #x, indices2 = self.pool2(x) 
      #      print(x.shape)
      #  print("dense input", self.dense_input) 
      #  print(" input height", self.input_height) 
        x = torch.flatten(x, 1) 
      #  print(x.shape)
        x = self.fc1(x)
        
       # x= self.md(x)
        if self.fc2_layer == 1:
            x = self.fc2(x)
        
        return x, indices1, indices2, size1, size2, size_conv1, size_conv2

    def decode(self, x, indices1, indices2, size1, size2, size_conv1, size_conv2):
        
        if self.fc2_layer ==1:
            x = F.relu(self.fc2_D(x))
        elif self.fc2_layer ==2:
            x = self.decodinglayer(x)
            
        x = F.relu(self.fc1_D(x))
    #    x = self.md(x)                                                ############DROPOUT
   #     print("before unflat", x.shape)
        x = self.unflatten(x)
  #      print("after unflat", x.shape)
        if self.num_conv_layers == 2:
            x=x
            x = self.unpool2_D(x, indices2, output_size=size2)
   #         print(x.shape)
        if self.num_conv_layers == 2:
            x = self.md(x)                                                ############DROPOUT
            x = F.relu(self.conv2_D(x, output_size=size_conv2))
            
        x = self.unpool_D(x, indices1, output_size=size1)
    #    print(x.shape)
        x = self.md(x)                                                ############DROPOUT
        x = self.conv1_D(x, output_size=size_conv1)                        #NO activation
#        print(x.shape)

        if self.act == 'softmax':
            #x = self.m(x)
            x = x   ### the softmax is incorporated in the CE loss 
        elif self.act == 'sigmoid':
            x = self.m(x)
        x = torch.squeeze(x, 1)
        return x

    def reparameterize(self, mu, log_var):
        """
        :param mu: mean from the encoder's latent space
        :param log_var: log variance from the encoder's latent space
        """
        std = torch.exp(0.5*log_var) # standard deviation
        eps = torch.randn_like(std) # `randn_like` as we need the same size
        sample = mu + (eps * std) # sampling as if coming from the input space
        return sample
    
    def forward(self, x, y_values=None):
      
        q, indices1, indices2, size1, size2, size_conv1, size_conv2 = self.encode(x)#.view(-1, 480)) 
        if self.fc2_layer ==1 or self.fc2_layer ==0:
            q_y = q.view(-1, self.categorical_dim)
            z = gumbel_softmax(q_y, self.temp, self.latent_dim, self.categorical_dim) 
            z = torch.reshape(z, (-1, self.latent_dim*self.categorical_dim))
            return self.decode(z, indices1, indices2, size1, size2, size_conv1, size_conv2),  F.softmax(q_y, dim=1)    ##################
        
        elif self.fc2_layer ==2:
            ## VANILLA VAE
            # get mu and log_var
            log_var = self.encodinglayer2_logvar(q)
            mean = self.encodinglayer2_mean(q)
            # get the latent vector through reparameterization
            z = self.reparameterize(mean, log_var)
            return self.decode(z, indices1, indices2, size1, size2, size_conv1, size_conv2), mean, log_var



#ANNEAL_RATE = 0.00000007 ### initial 0.00000007 -  change it for the test


# Reconstruction + KL divergence losses summed over all elements and batch
def loss_function(recon_x, x, qy, latent_dim, categorical_dim, beta, length, batch_size, act, reduction='sum'):
  #  print(x.shape)
    size = recon_x.shape[0]
    if act == 'sigmoid':
        BCE = F.binary_cross_entropy(recon_x.view(-1, length*4), x.view(-1, length*4), reduction=reduction)
    elif act == 'softmax':
    #    print(recon_x.shape)   #[batch, 4, length]
        recon_x = torch.transpose(recon_x, 2, 1) #[batch, length, 4]
        x = torch.transpose(x, 2, 1)         
        recon_x = torch.reshape(recon_x, (-1, 4))
        x = torch.reshape(x, (-1, 4))
               
        loss = nn.CrossEntropyLoss(reduction='sum')    
        BCE = loss(recon_x, torch.argmax(x, dim=1)) ######################## check if its correct the reshape
    
  #  log_ratio = torch.log(qy * categorical_dim + 1e-20)
 
 #   KLD = torch.sum(qy * log_ratio, dim=1)#.mean()
    
    log_qy = torch.log(qy * categorical_dim + 1e-20)   
    KLD_temp = qy*log_qy 
    KLD_temp = torch.reshape(KLD_temp, (-1, latent_dim, categorical_dim))  
  #  print(KLD_temp)

    KLD = torch.sum(KLD_temp, dim=(1,2))
    
    if reduction == 'none':
        if act == 'softmax':
            BCE = torch.reshape(BCE, (-1, length))
        BCE = torch.sum(BCE, dim=1)
    else:
        KLD = torch.sum(KLD)
        
   # beta = 1    
    if reduction == 'none':
        return ELBO, BCE, KLD 
    else:
        #ELBO = ELBO/size
        BCE = BCE/size
        KLD = KLD/size
        return BCE + beta*KLD, BCE, KLD

def loss_function_vanilla(recon_x, x, mu, log_var, length, beta, reduction='sum'):
    size = recon_x.shape[0]
    BCE = F.binary_cross_entropy(recon_x.view(-1, length*4), x.view(-1, length*4), reduction=reduction)
    KLD = -0.5 * (1 + log_var - mu.pow(2) - log_var.exp())
    if reduction == 'none':
        BCE = torch.sum(BCE, dim=1)
        KLD = torch.sum(KLD, dim=1)
    else:
        KLD = torch.sum(KLD)
        
   # beta = 1
    
    if reduction == 'none':
        return BCE + beta*KLD, BCE, KLD 
    else:

        BCE = BCE/size
        KLD = KLD/size
        return BCE + beta*KLD, BCE, KLD

def train(epoch, model, train_loader, lr, args):
    model.train()
    train_loss = 0
    
    optimizer = optim.Adam(model.parameters(), lr=lr)
    index = -1
    for batch_idx, (data, _) in enumerate(train_loader):
        
        index += 1
        # KL-annealing
        #beta = min(0.01 + ((epoch - 1) * 500 + index + 1) / 12500 , 1)
   #     beta  = beta = min(0.01 + (((epoch - 1)%25) * 4100  + index + 1) / 51250 , 1) ##cyclical KL annealing for batch size 128 for sliding dataset
  #      beta  = beta = min(0.01 + (((epoch - 1)%25) * 8200  + index + 1) / 51250 , 1) ##cyclical KL annealing for batch size 64 for 
      #  beta  = beta = min(0.01 + (((epoch - 1)%25) * 2050  + index + 1) / 25625 , 1) ##cyclical KL annealing for batch size 256 for
                                                                                    #12 epochs rising/12 epochs stable beta=1
    #    beta = min(0.01 + (((epoch - 1)%12) * 2050  + index + 1) / 12300 , 1) ##cyclical KL annealing for batch size 256 for
                                                                                    #6 epochs rising/6 epochs stable beta=1
        beta = min(0.01 + (((epoch - 1)) * 2050  + index + 1) / 51250 , 1) ##cyclical KL annealing for batch size 256 for
                                                                                    #6 epochs rising/6 epochs stable beta=1
     #   beta  = beta = min(0.01 + (((epoch - 1)%6) * 2050  + index + 1) / 6150 , 1) ##cyclical KL annealing for batch size 256 for
                                                                                    #12 epochs rising/12 epochs stable beta=1
     #   beta=1    
     #   
        #beta  = beta = min(0.01 + (((epoch - 1)) * 2050  + index + 1) / 102500 , 1) 
     #   print("beta", beta)
        data = Variable(data)
        if args.cuda:
            data = data.cuda()
        optimizer.zero_grad()
        
        #vanilla VAE or gumbel softmax VAE
        if model.fc2_layer ==2:
            recon_batch, mu, var  = model(data,model.temp)
            
            loss, bce, kl = loss_function_vanilla(recon_batch, data, mu, var, model.length, beta)
        else:
            recon_batch, qy  = model(data,model.temp)
      #      print(data.shape)
      #      print(recon_batch.shape)
            loss, bce, kl = loss_function(recon_batch, data, qy, model.latent_dim, model.categorical_dim, beta, model.length, args.batch_size, model.act)
            
        loss.backward()
        train_loss += loss.item()
        optimizer.step()
        
        if batch_idx % 100 == 1:
            model.temp = np.maximum(model.temp*np.exp(-model.temp_rate*batch_idx),model.temp_min)
        if batch_idx % args.log_interval == 0:
              print('Epoch: {:3d} [{:5d}/{:5d} ({:2.0f}%)]  \tLoss: {:11.6f}\trec: {:11.6f}\tkl: {:11.6f}'.format(
                    epoch, batch_idx * len(data), len(train_loader.dataset), 100. * batch_idx / len(train_loader),
                    loss.item(), bce.item(), kl.item()))  ####average loss per sample (average over the current batch)
            
    print('====> Epoch: {} Average loss: {:.4f}'.format(
          epoch, train_loss*args.batch_size/ len(train_loader.dataset)))
    print("temp", model.temp)
    print("beta", beta)
    
def test(epoch, model, test_loader, args):
    model.eval()
    test_loss = 0
    
    if args.test_model == True:
        loss_array = np.zeros(len(test_loader))
    index = -1    
        
    for i, (data, _) in enumerate(test_loader):
        index += 1 
        if args.cuda:
            data = data.cuda()
        data = Variable(data, volatile=True)
        
        #vanilla VAE or gumbel softmax VAE
        if model.fc2_layer ==2:
            recon_batch, mu, var  = model(data,model.temp)
            loss, bce, kl = loss_function_vanilla(recon_batch, data, mu, var, model.length, beta=1)
        else:
            recon_batch, qy  = model(data,model.temp)
            loss, bce, kl = loss_function(recon_batch, data, qy, model.latent_dim, model.categorical_dim, beta=1, length=model.length, batch_size=args.batch_size, act=model.act)            
            
       # loss, bce, kl = loss_function(recon_batch, data, qy, model.latent_dim, model.categorical_dim, beta=1)
        test_loss += loss.item()
        
        if args.test_model == True:
            loss_array[index] = batch_loss.item()
            
        if i % 100 == 1:
            model.temp = np.maximum(model.temp*np.exp(-model.temp_rate*i),model.temp_min)
        if i == 0:
            n = min(data.size(0), 8)

    test_loss /= len(test_loader.dataset)
    print('====> Test set loss: {:.4f}'.format(args.batch_size*test_loss))
    if args.test_model == True:
        return loss_array

def evaluate(test1_loader, model, args):
    model.eval()
    
    loss_array = np.zeros(len(test1_loader)*args.batch_size)
    index = -1    
        
    for i, (data, _) in enumerate(test1_loader):
        index += 1 
        if args.cuda:
            data = data.cuda()
        data = Variable(data, volatile=True)
        
        #vanilla VAE or gumbel softmax VAE
        if model.fc2_layer ==2:
            recon_batch, mu, var  = model(data,model.temp)

            loss, bce, kl = loss_function_vanilla(recon_batch, data, mu, var, model.length, beta=1, reduction='none')
        else:
            recon_batch, qy  = model(data,model.temp)
            loss, bce, kl = loss_function(recon_batch, data, qy, model.latent_dim, model.categorical_dim, beta=1, length=model.length, batch_size=args.batch_size, act=model.act, reduction='none')
        
        bce_ = bce.cpu().detach().numpy()
        loss_array[index*args.batch_size:(index*args.batch_size+bce_.shape[0])] = bce_
            
        if i % 100 == 1:
            model.temp = np.maximum(model.temp*np.exp(-model.temp_rate*i),model.temp_min)
        if i == 0:
            n = min(data.size(0), 8)

    return loss_array
    
def latent_repr(data_loader, model, args, testing=False):
    
    model.eval()   
    index = -1 
    latent = np.zeros((len(data_loader)*args.batch_size, model.latent_dim*model.categorical_dim))
   
    for data, _ in data_loader:  
        index += 1
        if args.cuda:
            data = data.cuda()
        length = model.length
        data = data.view(-1, 4, length) 
        
        a, i1, i2, size1, size2, size_conv1, size_conv2 = model.encode(data)
      #  a = a.view(-1, model.categorical_dim)
      #  z = gumbel_softmax(a, model.temp, model.latent_dim, model.categorical_dim) 
      #  z = torch.reshape(z, (-1, model.latent_dim*model.categorical_dim))
      # a=z
        if model.fc2_layer ==2:
            log_var = model.encodinglayer2_logvar(a)
            mean = model.encodinglayer2_mean(a)
            a = model.reparameterize(mean, log_var)

        latent[index*args.batch_size:((index+1)*args.batch_size), : ] = a.cpu().detach().numpy()
  
        
    return latent 

def reconstruction(data_loader, model):
    
    model.eval() 
    rec = np.zeros((len(data_loader), 4, 80))
    original = np.zeros((len(data_loader), 4, 80))
    index = -1
 
    for data, _ in data_loader:     
        
        index += 1
        
        if args.cuda:
            data = data.cuda()

        data = data.view(-1, 4, 80)  
        #Pass throught the VAE  
        recon_batch, qy = model(data, model.temp)
        
        rec[index, :, :] = recon_batch.detach().numpy()
        original[index, :, :] = data.detach().numpy()
        
        #############
        q, indices1, indices2, size1, size2, size_conv1, size_conv2 = model.encode(data)#.view(-1, 480))
        
        q_y = q.view(q.size(0), model.latent_dim, model.categorical_dim)
        z = gumbel_softmax(q_y, model.temp, model.latent_dim, model.categorical_dim)
        q_y =  torch.reshape(q, (q.size(0)* model.latent_dim, model.categorical_dim))      ##############################
        decoded = model.decode(z, indices1, indices2, size1, size2, size_conv1, size_conv2)
    return rec, original, q, z, decoded

def run(args):
    latent_dim = 4
    categorical_dim = 2
    model = VAE_gumbel_hp_tuning(0.7, latent_dim, categorical_dim, 12, 8, 3, 1, 2, 2)
    if args.cuda:
        model.cuda()
        print("Model on GPU")
    
    print("latent dim", latent_dim)
    
    #### Load data for the VAE training ####
  #  pos_seq_encoded_chr19 = loadtxt('chr19.fa_positive_80.csv', delimiter=',')  
  #  neg_seq_encoded_chr19 = loadtxt('chr19.fa_negative_80_uscs.csv', delimiter=',')
    
    pos_seq_encoded_chr19 = loadtxt('chr19.fa_positive_80.csv', delimiter=',')  
    pos_seq_encoded_chr21 = loadtxt('chr21.fa_positive_80.csv', delimiter=',')
    pos_seq_encoded_chr22 = loadtxt('chr22.fa_positive_80.csv', delimiter=',')
    pos_seq_encoded_chr7 = loadtxt('chr7.fa_positive_80.csv', delimiter=',')
    neg_seq_encoded_chr19 = loadtxt('chr19.fa_negative_80_uscs.csv', delimiter=',')
   # neg_seq_encoded_chr7 = loadtxt('chr7.fa_negative_80.csv', delimiter=',')

    #Concatenate positive and negative data 
    pos_seq_encoded_total = np.append(pos_seq_encoded_chr19, pos_seq_encoded_chr21, axis = 0)
    pos_seq_encoded_total = np.append(pos_seq_encoded_total, pos_seq_encoded_chr22, axis = 0)
    pos_seq_encoded_total = np.append(pos_seq_encoded_total, pos_seq_encoded_chr7, axis = 0)
    neg_seq_encoded_total =  neg_seq_encoded_chr19 #np.append(neg_seq_encoded_chr19, neg_seq_encoded_chr7, axis = 0)
    
    
 #   pos_seq_encoded_chr19 = loadtxt('chr19.test_positive_spikes.csv', delimiter=',')  
 #   pos_seq_encoded_chr21 = loadtxt('chr19.test_positive_spikes2.csv', delimiter=',')
 #   neg_seq_encoded_chr19 = loadtxt('chr19.test_negative_spikes.csv', delimiter=',')
   # neg_seq_encoded_chr7 = loadtxt('chr7.fa_negative_80.csv', delimiter=',')

    #Concatenate positive and negative data 
    pos_seq_encoded_total = pos_seq_encoded_total #np.append(pos_seq_encoded_chr19, pos_seq_encoded_chr21, axis = 0)
    neg_seq_encoded_total = neg_seq_encoded_total #np.append(neg_seq_encoded_chr19, neg_seq_encoded_chr7, axis = 0)
    
    print("negative", neg_seq_encoded_total.shape)
    print("positive", pos_seq_encoded_total.shape)
    
    data =  np.append(pos_seq_encoded_total, neg_seq_encoded_total, axis = 0)  #pos_seq_encoded_total
    train_loader, val_loader, test_loader, args = load_dataset(args, data)
    
    for epoch in range(1, args.epochs + 1):

        train(epoch, model, train_loader)
        test(epoch, model, test_loader)
        if args.cuda:
            sample = sample.cuda()

    json_dir = f"model.pt"
    print("Saving data at: " + json_dir)
    torch.save(model.state_dict(), json_dir)

if __name__ == '__main__':
    run(args)
